﻿Imports System.Runtime.InteropServices

Public Class USBListener
    Inherits NativeWindow

    Public Event DeviceChangeNotifyEvent As EventHandler(Of DeviceChangeEventArgs)
    Public Event TimerElapsed As EventHandler
    Private WithEvents SH_Timer As Timer

    Private Const S_OK As Integer = &H0
    Private Const WM_DESTROY As Integer = &H2
    Private Const WM_SHNOTIFY As Integer = &H488
    Private Const CSIDL_DESKTOP As Integer = &H0
    Private Const GPFIDL_DEFAULT As Integer = &H0

    Private m_hSHNotify As Integer = 0
    Private m_hNotifyLock As IntPtr = IntPtr.Zero

    <Flags>
    Enum SHChangeNotifyEventID As UInteger
        SHCNE_ALLEVENTS = &H7FFFFFFFL
        SHCNE_ASSOCCHANGED = &H8000000L
        SHCNE_ATTRIBUTES = &H800L
        SHCNE_CREATE = &H2L
        SHCNE_DELETE = &H4L
        SHCNE_DISKEVENTS = &H2381FL
        SHCNE_DRIVEADD = &H100L
        SHCNE_DRIVEADDGUI = &H10000L
        SHCNE_DRIVEREMOVED = &H80L
        SHCNE_EXTENDED_EVENT = &H4000000L
        SHCNE_FREESPACE = &H40000L
        SHCNE_GLOBALEVENTS = &HC0581E0L
        SHCNE_INTERRUPT = &H80000000L
        SHCNE_MEDIAINSERTED = &H20L
        SHCNE_MEDIAREMOVED = &H40L
        SHCNE_MKDIR = &H8L
        SHCNE_NETSHARE = &H200L
        SHCNE_NETUNSHARE = &H400L
        SHCNE_RENAMEFOLDER = &H20000L
        SHCNE_RENAMEITEM = &H1L
        SHCNE_RMDIR = &H10L
        SHCNE_SERVERDISCONNECT = &H4000L
        SHCNE_UPDATEDIR = &H1000L
        SHCNE_UPDATEIMAGE = &H8000L
        SHCNE_UPDATEITEM = &H2000L
    End Enum

    Private Enum SHCNRF As Integer
        SHCNRF_InterruptLevel = &H1
        SHCNRF_ShellLevel = &H2
        SHCNRF_RecursiveInterrupt = &H1000
        SHCNRF_NewDelivery = &H8000
    End Enum

    <StructLayout(LayoutKind.Sequential)>
    Private Structure SHChangeNotifyEntry
        Dim pidl As IntPtr
        <MarshalAs(UnmanagedType.Bool)> Dim fRecursive As Boolean
    End Structure

    <StructLayout(LayoutKind.Sequential)>
    Private Structure SHNOTIFYSTRUCT
        Dim dwItem1 As IntPtr
        Dim dwItem2 As IntPtr
    End Structure

    <DllImport("shell32.dll", EntryPoint:="SHChangeNotifyRegister")>
    Private Shared Function SHChangeNotifyRegister(
                            <[In]> ByVal hWnd As IntPtr,
                            <[In]> ByVal fSources As SHCNRF,
                            <[In]> ByVal fEvents As SHChangeNotifyEventID,'Integer,
                            <[In]> ByVal wMsg As Integer,
                            <[In]> ByVal cEntries As Integer,
                            <[In]> ByRef lpps As SHChangeNotifyEntry) As Integer
    End Function

    <DllImport("shell32.dll", EntryPoint:="SHChangeNotifyDeregister")>
    Private Shared Function SHChangeNotifyDeregister(<[In]> ByVal hNotify As Integer) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("shell32.dll", EntryPoint:="SHGetFolderLocation")>
    Private Shared Function SHGetFolderLocation(
                            <[In]> ByVal hwndOwner As IntPtr,
                            <[In]> ByVal nFolder As Integer,
                            <[In]> ByVal hToken As IntPtr,
                            <[In]> ByVal dwReserved As Integer,
                            <Out> ByRef pidl As IntPtr) As Integer
    End Function

    <DllImport("shell32.dll", EntryPoint:="SHChangeNotification_Lock")>
    Private Shared Function SHChangeNotification_Lock(
                            <[In]> ByVal hChange As IntPtr,
                            <[In]> ByVal dwProcId As Integer,
                            <Out> ByRef ppidl As IntPtr,
                            <Out> ByRef plEvent As SHChangeNotifyEventID) As IntPtr
    End Function

    <DllImport("shell32.dll", EntryPoint:="SHChangeNotification_Unlock")>
    Private Shared Function SHChangeNotification_Unlock(
                            <[In]> ByVal hLock As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("shell32.dll", EntryPoint:="SHGetPathFromIDListEx")>
    Private Shared Function SHGetPathFromIDListEx(
                            <[In]> ByVal pidl As IntPtr,
                            <[In]> <MarshalAs(UnmanagedType.LPWStr)> ByVal pszPath As String,
                            <[In]> ByVal cchPath As Integer,
                            <[In]> ByVal uOpts As Integer) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    Public Sub StartMonitoring()
        SH_Timer = New Timer With {.Interval = 3000, .Enabled = False}
        Dim params As New CreateParams()
        Me.CreateHandle(params)
        m_hSHNotify = RegisterNotify()
    End Sub
    Private Function RegisterNotify() As Integer
        Dim hSHNotify As Integer = 0
        Dim pPIDL As IntPtr = IntPtr.Zero
        Dim tSHCNE As New SHChangeNotifyEntry
        If SHGetFolderLocation(IntPtr.Zero, CSIDL_DESKTOP, IntPtr.Zero, 0, pPIDL) = S_OK Then
            With tSHCNE
                .pidl = pPIDL
                .fRecursive = True
            End With
            hSHNotify = SHChangeNotifyRegister(Me.Handle,
                                               SHCNRF.SHCNRF_ShellLevel Or
                                               SHCNRF.SHCNRF_InterruptLevel Or
                                               SHCNRF.SHCNRF_RecursiveInterrupt Or
                                               SHCNRF.SHCNRF_NewDelivery,
                                               SHChangeNotifyEventID.SHCNE_ALLEVENTS,
                                            WM_SHNOTIFY, 1, tSHCNE)
            Marshal.FreeCoTaskMem(pPIDL)
        End If
        Return hSHNotify
    End Function
    Protected Overrides Sub WndProc(ByRef m As Message)
        Select Case m.Msg
            Case WM_SHNOTIFY
                ProcessShellNotify(m)
            Case WM_DESTROY
                UnregisterNotify()
        End Select
        MyBase.WndProc(m)
    End Sub
    Private Sub ProcessShellNotify(ByVal m As Message)
        Dim intEvent As SHChangeNotifyEventID = 0
        Dim pPIDL As IntPtr = IntPtr.Zero
        If m_hNotifyLock = IntPtr.Zero Then
            m_hNotifyLock = SHChangeNotification_Lock(m.WParam, m.LParam.ToInt32, pPIDL, intEvent)
            If m_hNotifyLock <> IntPtr.Zero Then
                Dim tSHNS As SHNOTIFYSTRUCT = CType(Marshal.PtrToStructure(pPIDL, GetType(SHNOTIFYSTRUCT)), SHNOTIFYSTRUCT)
                If SHChangeNotification_Unlock(m_hNotifyLock) Then
                    m_hNotifyLock = IntPtr.Zero
                End If
                ShellNotifyEvent(intEvent, tSHNS)
            End If
        End If
    End Sub
    Private Sub UnregisterNotify()
        If m_hSHNotify <> 0 Then
            If SHChangeNotifyDeregister(m_hSHNotify) Then
                m_hSHNotify = 0
            End If
        End If
    End Sub
    Private Sub ShellNotifyEvent(ByVal intEvent As SHChangeNotifyEventID, ByVal tSHNS As SHNOTIFYSTRUCT)
        'Debug.WriteLine($"...............{intEvent}")
        If intEvent.HasFlag(SHChangeNotifyEventID.SHCNE_DRIVEADD) OrElse
            intEvent.HasFlag(SHChangeNotifyEventID.SHCNE_DRIVEREMOVED) OrElse
            intEvent.HasFlag(SHChangeNotifyEventID.SHCNE_MEDIAINSERTED) OrElse
            intEvent.HasFlag(SHChangeNotifyEventID.SHCNE_MEDIAREMOVED) Then
            Dim drivestate = intEvent.HasFlag(SHChangeNotifyEventID.SHCNE_DRIVEADD) OrElse intEvent.HasFlag(SHChangeNotifyEventID.SHCNE_MEDIAINSERTED)
            Dim strNullChar As Char = Convert.ToChar(0)
            Dim strPath1 As New String(strNullChar, 255)
            Dim strPath2 As New String(strNullChar, 255)
            If tSHNS.dwItem1 <> IntPtr.Zero Then
                If SHGetPathFromIDListEx(tSHNS.dwItem1, strPath1, strPath1.Length, GPFIDL_DEFAULT) Then
                    RaiseEvent DeviceChangeNotifyEvent(Me, New DeviceChangeEventArgs(drivestate))
                    SH_Timer.Start()
                End If
            End If
            If tSHNS.dwItem2 <> IntPtr.Zero Then
                If SHGetPathFromIDListEx(tSHNS.dwItem2, strPath2, strPath2.Length, GPFIDL_DEFAULT) Then
                    RaiseEvent DeviceChangeNotifyEvent(Me, New DeviceChangeEventArgs(drivestate))
                    SH_Timer.Start()
                End If
            End If
        End If
    End Sub
    Private Sub Timer_Tick(sender As Object, e As EventArgs) Handles SH_Timer.Tick
        SH_Timer.Stop()
        RaiseEvent TimerElapsed(Me, EventArgs.Empty)
    End Sub
End Class
Public Class DeviceChangeEventArgs
    Inherits EventArgs
    Public ReadOnly Property IsInserted As Boolean
    Public Sub New(isInserted As Boolean)
        Me.IsInserted = isInserted
    End Sub
End Class